/*
 * @(#)HParentLevel.java
 * 
 * Christopher Deckers (chrriis@brainlex.com)
 * http://chrriis.brainlex.com
 * 
 * See the file "LICENSE.txt" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package chrriis.uihierarchy;

import java.awt.Container;
import java.awt.LayoutManager;

import chrriis.uihierarchy.compatibility.Component;
import chrriis.uihierarchy.compatibility.SwingUtilities;
import chrriis.uihierarchy.compatibility.ArrayList;
import chrriis.uihierarchy.compatibility.List;

import chrriis.uihierarchy.constraints.IllegalConstraintsException;
import chrriis.uihierarchy.constraints.LayoutHConstraints;
import chrriis.uihierarchy.layout.HLayoutManager;

/**
 * A parent level that contains some children.
 * @version 1.0 2003.11.03
 * @author Christopher Deckers (chrriis@brainlex.com)
 */
public abstract class HParentLevel extends HLevel {

  /**
   * Construct a hierarchy parent level.
   * @param uihConfig The configuration.
   * @param parent The hierarchy parent level of this one.
   * @param component The component.
   */
  protected HParentLevel(UIHConfig uihConfig, HParentLevel parent, Object component) {
    super(uihConfig, parent, component);
  }

  /**
   * Get the container at this hierarchy level.
   * @return The container.
   */
  public Container getContainer() {
    return (Container)getComponent();
  }

  /** The list of children. */
  protected List childrenList = new ArrayList(5);
  /** The indexes of the children. */
  protected List indexesList = new ArrayList(5);

  /**
   * Get the list of children.
   * @return The list of children.
   */
  protected List getChildrenList() {
    return childrenList;
  }

  /**
   * Add a child level to the list of children.
   * @param hLevel The child level to add.
   */
  protected void addChild(HLevel hLevel) {
    addChild(-1, hLevel);
  }

  /**
   * Add a child level to the list of children at the specified index.
   * @param index The index at which to add the child.
   * @param hLevel The child level to add.
   */
  protected void addChild(int index, HLevel hLevel) {
    HLevel tmpHLevel = this;
    do {
      if(tmpHLevel.getComponent() == hLevel.getComponent()) {
        throw new UIHException(getConfig().createAccessor(hLevel), "A level cannot be added to itself or one of its own sub levels!");
      }
    } while((tmpHLevel = tmpHLevel.getParent()) != null);
    for(int i=0; i<childrenList.size(); i++) {
      HLevel level = (HLevel)childrenList.get(i);
      if(level.getComponent() == hLevel.getComponent()) {
        removeChild(level);
        break;
      }
    }
    hLevel.remove();
    hLevel.parent = this;
    childrenList.add(hLevel);
    indexesList.add(new Integer(index));
    if(isRealized()) {
      realizeChild(hLevel);
    }
  }

  /**
   * Remove a child level from the list of children.
   * @param hLevel The child level to remove.
   */
  protected void removeChild(final HLevel hLevel) {
    Container parentContainer = getContainer();
    if(parentContainer != null && Component.isDisplayable(parentContainer) && !SwingUtilities.isEventDispatchThread()) {
      try {
        SwingUtilities.invokeAndWait(new Runnable() {
          public void run() {
            removeChild(hLevel);
          }
        });
      } catch(Exception e) {
        e.printStackTrace();
      }
      return;
    }
    int index = childrenList.indexOf(hLevel);
    if(index != -1) {
      indexesList.remove(index);
      childrenList.remove(index);
      hLevel.parent = null;
      if(parentContainer != null && isRealized()) {
        parentContainer.remove(hLevel.getComponent());
        parentContainer.validate();
        parentContainer.repaint();
      }
    }
  }

/*----------------------------------------------------------------------------*/

  /** The layout manager in use. Null is a possible layout manager. */
  protected LayoutManager layoutManager;

  /** Indicate if the layout manager is defined by the hierarchy. */
  protected boolean isLayoutManagerDefined = false;

  /**
   * Indicate whether the layout manager is defined.
   * @return True if the layout manager is defined.
   */
  protected boolean isLayoutManagerDefined() {
    return isLayoutManagerDefined;
  }

  /**
   * Use the specified layout manager.
   * @param layoutManager The layout manager to use.
   */
  protected void layout_(HLayoutManager hLayoutManager) {
    layout_(hLayoutManager.createLayout(getConfig()));
  }

  /**
   * Use the specified layout manager.
   * @param layoutManager The layout manager to use.
   */
  protected void layout_(LayoutManager layoutManager) {
    if(isLayoutManagerDefined) {
      throw new UIHException(getConfig().createAccessor(this), "The layout method cannot be called twice on the same object!");
    }
    this.layoutManager = layoutManager;
    isLayoutManagerDefined = true;
  }

  /**
   * Get the layout manager in use.
   * @return The layout manager. Null is a possible value.
   */
  protected LayoutManager getLayoutManager() {
    if(isLayoutManagerDefined) {
      return layoutManager;
    }
    return getContainer().getLayout();
  }

/*----------------------------------------------------------------------------*/

  /** The sub constraints, if any. */
  protected Object subConstraints;

  /**
   * Get the sub constraints.
   * @return The sub constraints.
   */
  protected Object getSubConstraints() {
    return subConstraints;
  }

  /**
   * Set the sub constraints.
   * @param subConstraints The sub constraints.
   */
  protected void subConstrain_(Object subConstraints) {
    if(this.subConstraints != null) {
      throw new UIHException(getConfig().createAccessor(this), "The subConstrain method cannot be called twice on the same object!");
    }
    this.subConstraints = subConstraints;
  }

/*----------------------------------------------------------------------------*/

  /**
   * Realize the level.
   */
  protected void realize() {
    if(isRealized()) {
      return;
    }
    super.realize();
    Container parentContainer = getContainer();
    if(parentContainer != null && Component.isDisplayable(parentContainer) && !SwingUtilities.isEventDispatchThread()) {
      try {
        SwingUtilities.invokeAndWait(new Runnable() {
          public void run() {
            doRealization();
          }
        });
      } catch(Exception e) {
        e.printStackTrace();
      }
      return;
    }
    doRealization();
  }

  /**
   * Realize the level.
   */
  protected void doRealization() {
    Container parentContainer = getContainer();
    if(isLayoutManagerDefined()) {
      parentContainer.setLayout(getLayoutManager());
    }
    List childrenList = getChildrenList();
    for(int i=0; i<childrenList.size(); i++) {
      HLevel child = (HLevel)childrenList.get(i);
      realizeChild(child);
    }
    Object subConstraints = getSubConstraints();
    if(subConstraints != null) {
      LayoutHConstraints layoutHConstraints;
      if(subConstraints instanceof LayoutHConstraints) {
        layoutHConstraints = (LayoutHConstraints)subConstraints;
      } else {
        LayoutManager parentLayout = parentContainer.getLayout();
        if(parentLayout == null) {
          layoutHConstraints = getConfig().getLayoutConstraints(null);
        } else {
          layoutHConstraints = getConfig().getLayoutConstraints(parentLayout.getClass());
        }
      }
      if(layoutHConstraints != null) {
        SingleLevelAccessor childAccessor = getConfig().createAccessor(this);
        try {
          layoutHConstraints.createConstraints(childAccessor, parentContainer, subConstraints);
        } catch(Exception e) {
          throw new IllegalConstraintsException(childAccessor, parentContainer, constraints, e);
        }
      }
    }
  }

  /**
   * Realize a child of this parent component. This method realizes the child,
   * computes all the constraints and add the child. 
   * @param hLevel The child level to realize.
   */
  protected void realizeChild(final HLevel hLevel) {
    Container parentContainer = getContainer();
    if(parentContainer != null && Component.isDisplayable(parentContainer) && !SwingUtilities.isEventDispatchThread()) {
      try {
        SwingUtilities.invokeAndWait(new Runnable() {
          public void run() {
            realizeChild(hLevel);
          }
        });
      } catch(Exception e) {
        e.printStackTrace();
      }
      return;
    }
    hLevel.realize();
    if(parentContainer != null && !(hLevel instanceof HRootNode)) {
      Object constraints = hLevel.getConstraints();
      LayoutHConstraints layoutHConstraints;
      if(constraints instanceof LayoutHConstraints) {
        layoutHConstraints = (LayoutHConstraints)constraints;
      } else {
        LayoutManager parentLayout = parentContainer.getLayout();
        if(parentLayout == null) {
          layoutHConstraints = getConfig().getLayoutConstraints(null);
        } else {
          layoutHConstraints = getConfig().getLayoutConstraints(parentLayout.getClass());
        }
      }
      if(layoutHConstraints != null) {
        SingleLevelAccessor childAccessor = getConfig().createAccessor(hLevel);
        try {
          constraints = layoutHConstraints.createConstraints(childAccessor, parentContainer, constraints);
        } catch(Exception e) {
          throw new IllegalConstraintsException(childAccessor, parentContainer, constraints, e);
        }
      }
      try {
        int index = ((Integer)indexesList.get(getChildrenList().indexOf(hLevel))).intValue();
        java.awt.Component component = hLevel.getComponent();
        if(constraints == null) {
          if(index == -1) {
            parentContainer.add(component);
          } else {
            parentContainer.add(component, index);
          }
        } else {
          if(index == -1) {
            parentContainer.add(component, constraints);
          } else {
            parentContainer.add(component, constraints, index);
          }
        }
        parentContainer.validate();
        parentContainer.repaint();
      } catch(Exception e) {
        throw new UIHException(getConfig().createAccessor(hLevel), "Failed to add level to parent!", e);
      }
    }
  }

}
