/*
 * @(#)UIHConfig.java
 * 
 * Christopher Deckers (chrriis@brainlex.com)
 * http://chrriis.brainlex.com
 * 
 * See the file "LICENSE.txt" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package chrriis.uihierarchy;

import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.GridBagLayout;
import java.util.StringTokenizer;

import chrriis.uihierarchy.accessor.AccessorFactory;
import chrriis.uihierarchy.compatibility.ArrayList;
import chrriis.uihierarchy.compatibility.Collection;
import chrriis.uihierarchy.compatibility.Compatibility;
import chrriis.uihierarchy.compatibility.HashMap;
import chrriis.uihierarchy.compatibility.HashSet;
import chrriis.uihierarchy.compatibility.List;
import chrriis.uihierarchy.compatibility.Map;
import chrriis.uihierarchy.compatibility.Set;
import chrriis.uihierarchy.compatibility.SwingUtilities;
import chrriis.uihierarchy.constraints.BorderHConstraints;
import chrriis.uihierarchy.constraints.BoxHConstraints;
import chrriis.uihierarchy.constraints.FlowHConstraints;
import chrriis.uihierarchy.constraints.GridBagHConstraints;
import chrriis.uihierarchy.constraints.LayoutHConstraints;
import chrriis.uihierarchy.constraints.NullHConstraints;
import chrriis.uihierarchy.constraints.SpringHConstraints;
import chrriis.uihierarchy.creator.ComponentCreator;
import chrriis.uihierarchy.creator.NullComponentCreator;
import chrriis.uihierarchy.creator.StringComponentCreator;
import chrriis.uihierarchy.debug.Debugger;
import chrriis.uihierarchy.unit.Unit;

/**
 * The centralized place for UIH configuration. The configuration also retains
 * the information regarding names to components mapping.
 * @version 1.0 2003.12.2
 * @author Christopher Deckers (chrriis@brainlex.com)
 */
public class UIHConfig {

  /**
   * Construct a UIHConfig.
   */
  protected UIHConfig() {
    UIHConfig config = getDefault();
    if(config != null && config != this) {
      setAutoRealizing(config.isAutoRealizing());
    }
  }

/*----------------------------------------------------------------------------*/

  /** The default configuration. */
  protected static UIHConfig defaultConfig = new UIHConfig();

  /**
   * Get the default configuration, which is used when nothing specific was
   * configured for a particular instance. Any change made to this default
   * configuration will affect all the configurations.
   * @return The default configuration.
   */
  public static UIHConfig getDefault() {
    return defaultConfig;
  }

  /** Indicate if Swing is preferred over AWT. */
  protected boolean isSwingPreferred = true;

  /**
   * Indicate if Swing is preferred over AWT.
   * @return True if Swing is preferred over AWT for component implicit
   * generation.
   */
  public boolean isSwingPreferred() {
    return Compatibility.isSwingPresent() && isSwingPreferred;
  }

  /**
   * Set whether Swing is preferred over AWT for implicit component generation.
   * This is true by defaul.
   * @param isSwingPreferred True if Swing should be used when available.
   */
  public void setSwingPreferred(boolean isSwingPreferred) {
    this.isSwingPreferred = isSwingPreferred;
  }

  static {
    UIHConfig defaultConfig = getDefault();
    // Set the auto resizing
    defaultConfig.setAutoRealizing(true);
    // Register default component creators
    defaultConfig.setDefaultComponentCreators();
    // Register default layout constraints
    defaultConfig.setDefaultLayoutConstraints();
    // Set the default accessor factory
    defaultConfig.setAccessorFactory(new AccessorFactory());
    // Set the default debugger
    defaultConfig.setDebugger(new Debugger());
  }

  /**
   * Register the default component creators.
   */
  protected void setDefaultComponentCreators() {
    setComponentCreator(null, new NullComponentCreator());
    setComponentCreator(String.class, new StringComponentCreator());
  }

  /**
   * Register the default layout constraints.
   */
  protected void setDefaultLayoutConstraints() {
    setLayoutConstraints(null, new NullHConstraints());
    setLayoutConstraints(FlowLayout.class, new FlowHConstraints());
    setLayoutConstraints(BorderLayout.class, new BorderHConstraints());
    setLayoutConstraints(GridBagLayout.class, new GridBagHConstraints());
    try {
      setLayoutConstraints(Class.forName("javax.swing.BoxLayout"), new BoxHConstraints());
      try {
        setLayoutConstraints(Class.forName("javax.swing.SpringLayout"), new SpringHConstraints());
      } catch(Throwable t) {}
    } catch(Throwable t) {}
    try {
      setLayoutConstraints(Class.forName("com.jgoodies.forms.layout.FormLayout"), (LayoutHConstraints)Class.forName("chrriis.uihierarchy.constraints.FormHConstraints").newInstance());
    } catch(Throwable t) {}
    try {
      setLayoutConstraints(Class.forName("com.zookitec.layout.ExplicitLayout"), (LayoutHConstraints)Class.forName("chrriis.uihierarchy.constraints.ExplicitHConstraints").newInstance());
    } catch(Throwable t) {}
  }

/*----------------------------------------------------------------------------*/

  /** Indicate if the hierarchy auto realizes itself at the last close(). */
  protected boolean isAutoRealizing = false;

  /**
   * Indicate if the hierarchy auto realizes itself at the last close().
   * @return True if the hierarchy auto realizes itself.
   */
  public boolean isAutoRealizing() {
    return isAutoRealizing;
  }

  /**
   * Set whether the hierarchy will auto realize itself at the last
   * <code>close()</code> or if an explicit call to <code>realize()</code> must
   * be made after the last <code>close()</code>.
   * @param isAutoRealizing True if the hierarchy is to realize itself
   * automatically.
   */
  public void setAutoRealizing(boolean isAutoRealizing) {
    this.isAutoRealizing = isAutoRealizing;
  }

/*----------------------------------------------------------------------------*/

  /** The map defining the default constraints associated to layouts. */
  protected Map layoutHConstraintsMap;

  /**
   * Register a layout constraints handler associated to a layout class.
   * @param layoutClass The layout class for which to set the constraints.
   * @param layoutHConstraints The constraints.
   */
  public void setLayoutConstraints(Class layoutClass, LayoutHConstraints layoutHConstraints) {
    if(layoutHConstraintsMap == null) {
      layoutHConstraintsMap = new HashMap();
    }
    layoutHConstraintsMap.put(layoutClass, layoutHConstraints);
  }

  /**
   * Get the layout constraints handler associated to a layout class.
   * @param layoutClass The layout class for which to set the constraints.
   * @return The constraints, or null if none are associated.
   */
  public LayoutHConstraints getLayoutConstraints(Class layoutClass) {
    UIHConfig defaultConfig = getDefault();
    if(layoutHConstraintsMap != null) {
      LayoutHConstraints layoutHConstraints = (LayoutHConstraints)layoutHConstraintsMap.get(layoutClass);
      if(layoutHConstraints != null || this == defaultConfig) {
        return layoutHConstraints;
      }
    }
    return defaultConfig.getLayoutConstraints(layoutClass);
  }

/*----------------------------------------------------------------------------*/

  /** The map of components */
  protected Map componentCreatorsMap;

  /**
   * Register a component creator associated to an object class.
   * @param objectClass The class of the object the creator handles.
   * @param componentCreator The creator of components to associate.
   */
  public void setComponentCreator(Class objectClass, ComponentCreator componentCreator) {
    if(componentCreatorsMap == null) {
      componentCreatorsMap = new HashMap();
    }
    componentCreatorsMap.put(objectClass, componentCreator);
  }

  /**
   * Get the component creator associated to an object class.
   * @param objectClass The class of the object the creator handles.
   * @return The component creator, or null if none is associated.
   */
  public ComponentCreator getComponentCreator(Class objectClass) {
    UIHConfig defaultConfig = getDefault();
    if(componentCreatorsMap != null) {
      ComponentCreator componentCreator = (ComponentCreator)componentCreatorsMap.get(objectClass);
      if(componentCreator != null || this == defaultConfig) {
        return componentCreator;
      }
    }
    return defaultConfig.getComponentCreator(objectClass);
  }

/*----------------------------------------------------------------------------*/

  /** The factory of accessors. */
  protected AccessorFactory accessorFactory;

  /**
   * Set the accessor factory to use when an accessor needs to be created.
   * @param accessorFactory The new factory of accessors.
   */
  public void setAccessorFactory(AccessorFactory accessorFactory) {
    this.accessorFactory = accessorFactory;
  }

  /**
   * Get the accessor factory, used to create the level accessors.
   * @return The accessor factory.
   */
  protected AccessorFactory getAccessorFactory() {
    if(accessorFactory != null) {
      return accessorFactory;
    }
    return getDefault().getAccessorFactory();
  }

/*----------------------------------------------------------------------------*/

  /** The map from names to HLevels. */
  protected Map namesToHLevelMap;
  
  /**
   * Map the given names to the current hierarchy level.
   * @param hLevel The hierarchy level.
   * @param names The names of the level, which are commas separated (leading
   * and trailing spaces are removed).
   */
  protected void map(HLevel hLevel, String names) {
    StringTokenizer st = new StringTokenizer(names, ",");
    if(st.countTokens() == 0) {
      throw new IllegalArgumentException("Cannot map to empty names!");
    }
    if(namesToHLevelMap == null) {
      namesToHLevelMap = new HashMap();
    }
    while(st.hasMoreTokens()) {
      String key = st.nextToken().trim();
      Object value = namesToHLevelMap.get(key);
      if(value == null) {
        namesToHLevelMap.put(key, hLevel);
      } else {
        if(!(value instanceof Set)) {
          Set valuesSet = new HashSet(2); 
          valuesSet.add(value);
          namesToHLevelMap.put(key, valuesSet);
          value = valuesSet;
        }
        ((Set)value).add(hLevel);
      }
    }
  }
  
  /**
   * Get the list of mapping keys of a hierarchy level. This method is not
   * really efficient and is mainly used during debugging (i.e. when printing
   * the hierarchy).
   * @param hLevel The hierarchy level for which to find the mappings.
   * @return The list of keys.
   */
  protected List getMapList(HLevel hLevel) {
    List keyList = new ArrayList();
    if(namesToHLevelMap != null) {
      Set keySet = namesToHLevelMap.keySet();
      Object[] keys = keySet.toArray();
      for(int i=0; i<keys.length; i++) {
        Object key = keys[i];
        Object value = namesToHLevelMap.get(key);
        if(value == hLevel || (value instanceof Set && ((Set)value).contains(hLevel))) {
          keyList.add(key);
        }
      }
    }
    return keyList;
  }

  /**
   * Get the object that is mapped to the specified name.
   * @param name The name that is mapped.
   * @return The mapped object.
   */
  protected Object getMapped(String name) {
    if(namesToHLevelMap == null) {
      return null; 
    }
    return namesToHLevelMap.get(name);
  }

/*----------------------------------------------------------------------------*/

  /**
   * The info representing a debug level.
   */
  protected class LevelDebuggingInfo {
    /** The levels to debug. */
    public HLevel[] hLevels;
    /** The parameters to use with the debugger. */
    public Object parameters;
    /**
     * Construct a level debugging info.
     * @param hLevels The levels to debug.
     * @param parameters The parameters to use with the debugger.
     */
    protected LevelDebuggingInfo(HLevel[] hLevels, Object parameters) {
      this.hLevels = hLevels;
      this.parameters = parameters;
    }
  }

  /** The list of debug info. */
  protected ArrayList levelDebuggingInfosList;

  /**
   * Add some levels to the debugging.
   * @param levelNames The names of the mapped levels to debug.
   * @param parameters The parameters to use with the debugger.
   */
  protected void addLevelsDebugging(String levelNames, Object parameters) {
    Collection mappedLevels = getMappedLevels(levelNames);
    if(mappedLevels.size() > 0) {
      HLevel[] hLevels = (HLevel[])mappedLevels.toArray(new HLevel[0]);
      if(hLevels[0].isRealized()) {
        debugLevels(hLevels, parameters);
      } else {
        if(levelDebuggingInfosList == null) {
          levelDebuggingInfosList = new ArrayList(1);
        }
        levelDebuggingInfosList.add(new LevelDebuggingInfo(hLevels, parameters));
      }
    }
  }

  /**
   * Add a level to the debugging.
   * @param hLevel The level to debug.
   * @param parameters The parameters to use with the debugger.
   */
  protected void addLevelDebugging(HLevel hLevel, Object parameters) {
    if(hLevel.isRealized()) {
      debugLevels(new HLevel[] {hLevel}, parameters);
    } else {
      if(levelDebuggingInfosList == null) {
        levelDebuggingInfosList = new ArrayList(1);
      }
      levelDebuggingInfosList.add(new LevelDebuggingInfo(new HLevel[] {hLevel}, parameters));
    }
  }

  /**
   * Process the levels debugging
   */
  protected void processLevelsDebugging() {
    if(levelDebuggingInfosList != null) {
      for(int i=0; i<levelDebuggingInfosList.size(); i++) {
        LevelDebuggingInfo info = (LevelDebuggingInfo)levelDebuggingInfosList.get(i);
        debugLevels(info.hLevels, info.parameters);
      }
      levelDebuggingInfosList = null;
    }
  }

  /**
   * Debug the levels.
   * @param hLevels The levels to debug.
   * @param parameters The parameters to use to debug, which can actually be a
   * debugger to use.
   */
  protected void debugLevels(final HLevel[] hLevels, final Object parameters) {
    if(!SwingUtilities.isEventDispatchThread()) {
      try {
        SwingUtilities.invokeAndWait(new Runnable() {
          public void run() {
            debugLevels(hLevels, parameters);
          }
        });
      } catch(Exception e) {
        e.printStackTrace();
      }
      return;
    }
    if(parameters instanceof Debugger) {
      ((Debugger)parameters).debug(createAccessor(hLevels), null);
    } else {
      getDebugger().debug(createAccessor(hLevels), parameters);
    }
  }

  /** The debugger. */
  protected Debugger debugger;

  /**
   * Set the debugger to use to debug the hierarchy levels.
   * @param debugger The new debugger.
   */
  public void setDebugger(Debugger debugger) {
    this.debugger = debugger;
  }

  /**
   * Get the debugger, used to debug the hierarchy levels.
   * @return The debugger.
   */
  protected Debugger getDebugger() {
    if(debugger != null) {
      return debugger;
    }
    return getDefault().getDebugger();
  }

/*----------------------------------------------------------------------------*/

  /**
   * Create an accessor for some mapped levels. If some levels appear more than
   * once in the mapping, they will be considered only once.<br>
   * This method uses the accessor factory currently defined, so one can cast
   * the resulting accessor to a more suitable accessor.
   * @param names The names of the hierarchy levels to get, in a comma separated
   * String if more than one name is to be specified.
   * @return The hierarchy level accessor.
   * @see chrriis.uihierarchy.accessor.AccessorFactory
   */
  public Accessor createAccessor(String names) {
    Collection hLevelsSet = getMappedLevels(names);
    if(hLevelsSet.size() == 0) {
      throw new IllegalArgumentException("Cannot create an accessor for empty names!");
    }
    return createAccessor((HLevel[])hLevelsSet.toArray(new HLevel[0]));
  }

  /**
   * Create an accessor for some mapped levels, with excluding some. If several
   * levels appear more than once in the mapping, they will be considered only
   * once.<br>
   * This method uses the accessor factory currently defined, so one can cast
   * the resulting accessor to a more suitable accessor.
   * @param names The names of the hierarchy levels to get, in a comma separated
   * String if more than one name is to be specified.
   * @param excludedNames The names of the levels to exclude when creating an
   * accessor on the names. 
   * @return The hierarchy level accessor.
   * @see chrriis.uihierarchy.accessor.AccessorFactory
   */
  public Accessor createAccessor(String names, String excludedNames) {
    Collection hLevelsSet = getMappedLevels(names);
    if(excludedNames != null) {
      hLevelsSet.removeAll(getMappedLevels(excludedNames));
    }
    if(hLevelsSet.size() == 0) {
      throw new IllegalArgumentException("Cannot create an accessor for empty names!");
    }
    return createAccessor((HLevel[])hLevelsSet.toArray(new HLevel[0]));
  }

  /**
   * Get the mapped levels. If some levels appear more than once in the mapping,
   * they will be considered only once.
   * @param names The names of the hierarchy levels to get, in a comma separated
   * String if more than one name is to be specified.
   * @return A collection of mapped levels.
   */
  protected Collection getMappedLevels(String names) {
    StringTokenizer st = new StringTokenizer(names, ",");
    Collection hLevelsSet = new HashSet();
    while(st.hasMoreTokens()) {
      String name = st.nextToken().trim();
      Object value = getMapped(name);
      if(value == null) {
        throw new IllegalArgumentException("Cannot find \"" + name + "\": the name is not mapped!");
      }
      if(value instanceof Set) {
        hLevelsSet.addAll((Set)value);
      } else {
        hLevelsSet.add(value);
      }
    }
    return hLevelsSet;
  }

  /**
   * Create a single level accessor using the accessor factory that is
   * registered.
   * @param hLevel The level for which to get the accessor.
   * @return The accessor that is created.
   */
  protected SingleLevelAccessor createAccessor(HLevel hLevel) {
    return (SingleLevelAccessor)getAccessorFactory().createAccessor(this, new HLevel[] {hLevel});
  }

  /**
   * Create an accessor using the accessor factory that is registered.
   * @param hLevels The levels for which to get the accessor.
   * @return The accessor that is created.
   */
  protected Accessor createAccessor(HLevel[] hLevels) {
    return getAccessorFactory().createAccessor(this, hLevels);
  }

/*----------------------------------------------------------------------------*/

  /** The map to associate some names to some units. */
  protected Map nameToUnitMap;

  /**
   * Register a unit, using the names it defines.
   * @param unit The unit to register.
   */
  public void registerUnit(Unit unit) {
    registerUnit(unit, unit.getNames());
  }

  /**
   * Register a unit, with some names that define it.
   * @param unit The unit to register.
   * @param names The names that identify the unit.
   */
  public void registerUnit(Unit unit, String[] names) {
    if(names == null || names.length == 0) {
      throw new IllegalStateException("The unit must have some identifying names!");
    }
    for(int i=0; i<names.length; i++) {
      String name = names[i];
      if(name == null || "".equals(name)) {
        throw new IllegalStateException("An identifying name of a unit cannot be null or empty!");
      }
      if(nameToUnitMap == null) {
        nameToUnitMap = new HashMap();
      }
      Unit oldUnit = (Unit)nameToUnitMap.get(name);
      if(oldUnit != null && !oldUnit.equals(unit)) {
        System.err.println("A unit with the name \"" + name + "\" is already defined! Skipping addition.");
      } else {
        nameToUnitMap.put(name, unit);
      }
    }
  }

  /**
   * Get the unit that is registered with the given name.
   * @param unitName The name of the unit to get.
   * @return The unit or null if it could not be found.
   */
  public Unit getUnit(String unitName) {
    UIHConfig defaultConfig = getDefault();
    if(nameToUnitMap != null) {
      Unit unit = (Unit)nameToUnitMap.get(unitName);
      if(unit != null) {
        return unit;
      }
    }
    if(this == defaultConfig) {
      return Unit.getUnit(unitName);
    }
    return defaultConfig.getUnit(unitName);
  }

}
