/*
 * @(#)DluY.java
 * 
 * Christopher Deckers (chrriis@brainlex.com)
 * http://chrriis.brainlex.com
 * 
 * See the file "LICENSE.txt" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package chrriis.uihierarchy.unit;

import java.awt.Button;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Toolkit;

import javax.swing.JButton;
import javax.swing.UIManager;

import chrriis.uihierarchy.compatibility.Compatibility;

/**
 * A handler to manipulate a default unit representing a dialog unit on the
 * vertical axis.
 * @version 1.0 2005.02.01
 * @author Christopher Deckers (chrriis@brainlex.com)
 */
public class DluY {
  
  /**
   * A unit representing a dialog unit on the vertical axis.
   * @version 1.0 2005.02.01
   * @author Christopher Deckers (chrriis@brainlex.com)
   */
  public static class DluYUnit extends Unit {

    /** The string that is used as a reference. */
    protected String referenceString = "X";
    
    /** The font to use as a reference. */
    protected Font referenceFont;
    
    /** The ratio. */
    protected double dluYUnitRatio;

    /**
     * Construct a new DluYUnit, using the given font.
     * @param referenceFont The font to use as a reference.
     */
    protected DluYUnit(Font referenceFont) {
      this.referenceFont = referenceFont;
      FontMetrics metrics = Toolkit.getDefaultToolkit().getFontMetrics(referenceFont);
      int ascent = metrics.getAscent();
      dluYUnitRatio = (ascent > 14 ? ascent : ascent + (15 - ascent) / 3.0) / 8.0;
    }

    /**
     * Get the ratio to use when making conversions. The ratio is so that:<br>
     * <code>unitValue * ratio == pixelValue</code>.
     * @return The ratio to use for conversions.
     */
    protected double getRatio() {
      return dluYUnitRatio;
    }

    /**
     * Get the double value resulting from the conversion of the value from the specified unit.
     * @param unit The unit from which to convert.
     * @param value The value to convert.
     * @return The result of the conversion.
     */
    public double doubleValue(Unit unit, double value) {
      return convert(unit, value);
    }

    /**
     * Get the names representing the unit.
     * @return An array containing the names.
     */
    public String[] getNames() {
      return new String[] {"duy", "dluy"};
    }
  }

  /** Hiden constructor of a DluY. */
  protected DluY() {}

  /** The default unit that is used. */
  protected static DluYUnit unit;
  
  /**
   * Get the default unit.
   * @return The default unit.
   */
  public static DluYUnit getUnit() {
    synchronized(DluY.class) {
      if(unit == null) {
        Font referenceFont;
        if(Compatibility.isSwingPresent()) {
          referenceFont = UIManager.getFont("Button.font");
          if(referenceFont == null) {
            referenceFont = new JButton().getFont();
          }
        } else {
          referenceFont = new Button().getFont();
        }
        unit = newInstance(referenceFont);
      }
    }
    return unit;
  }

  /**
   * Create a new instance of this unit, based on the given font.
   * @param referenceFont The font to use as a reference.
   * @return The newly created instance.
   */
  public static DluYUnit newInstance(Font referenceFont) {
    return new DluYUnit(referenceFont);
  }

  /**
   * Get the double value resulting from the conversion of the value from the specified unit.
   * @param unitName A name identifying the unit from which to convert.
   * @param value The value to convert.
   * @return The result of the conversion.
   */
  public static double doubleValue(String unitName, double value) {
    Unit unit = Unit.getUnit(unitName);
    if(unit == null) {
      throw new IllegalArgumentException("The unit \""+ unit + "\" is not registered!");
    }
    return getUnit().doubleValue(unit, value);
  }
  
  /**
   * Get the double value resulting from the conversion of the value from the specified unit.
   * @param unit The unit from which to convert.
   * @param value The value to convert.
   * @return The result of the conversion.
   */
  public static double doubleValue(Unit unit, double value) {
    return getUnit().doubleValue(unit, value);
  }

  /**
   * Parse a value along with its unit from a String representation, and
   * convert it to this unit.
   * @param unitValue The value with a unit in its String
   * representation. If the unit is not specified, then assume it is Pixel.
   * @return The int value resulting from the conversion.
   */
  public static int intValue(String unitValue) {
    return (int)Math.round(doubleValue(unitValue));
  }

  /**
   * Parse a value along with its unit from a String representation, and
   * convert it to this unit.
   * @param unitValue The value with a unit in its String
   * representation. If the unit is not specified, then assume it is Pixel.
   * @return The double value resulting from the conversion.
   */
  public static double doubleValue(String unitValue) {
    return doubleValue(Pixel.getUnit(), Unit.parse(unitValue));
  }

}
