/*
 * @(#)Inch.java
 * 
 * Christopher Deckers (chrriis@brainlex.com)
 * http://chrriis.brainlex.com
 * 
 * See the file "LICENSE.txt" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package chrriis.uihierarchy.unit;

/**
 * A handler to manipulate a default unit representing an inch.
 * @version 1.0 2005.02.01
 * @author Christopher Deckers (chrriis@brainlex.com)
 */
public class Inch {

  /**
   * A unit representing an inch.
   * @version 1.0 2005.02.01
   * @author Christopher Deckers (chrriis@brainlex.com)
   */
  public static class InchUnit extends ScreenUnit {
    
    /**
     * Construct a new InchUnit.
     */
    protected InchUnit() {
      super();
    }
    
    /**
     * Get the names representing the unit.
     * @return An array containing the names.
     */
    public String[] getNames() {
      return new String[] {"in", "inch"};
    }
  }
  
  /** Hiden constructor of an Inch. */
  protected Inch() {}
  
  /** The default unit that is used. */
  protected static InchUnit unit;
  
  /**
   * Get the default unit.
   * @return The default unit.
   */
  public static InchUnit getUnit() {
    synchronized(Inch.class) {
      if(unit == null) {
        unit = newInstance();
      }
    }
    return unit;
  }
  
  /**
   * Create a new instance of this unit.
   * @return The newly created instance.
   */
  public static InchUnit newInstance() {
    return new InchUnit();
  }

  /**
   * Get the double value resulting from the conversion of the value from the specified unit.
   * @param unitName A name identifying the unit from which to convert.
   * @param value The value to convert.
   * @return The result of the conversion.
   */
  public static double doubleValue(String unitName, double value) {
    Unit unit = Unit.getUnit(unitName);
    if(unit == null) {
      throw new IllegalArgumentException("The unit \""+ unit + "\" is not registered!");
    }
    return getUnit().doubleValue(unit, value);
  }
  
  /**
   * Get the double value resulting from the conversion of the value from the specified unit.
   * @param unit The unit from which to convert.
   * @param value The value to convert.
   * @return The result of the conversion.
   */
  public static double doubleValue(Unit unit, double value) {
    return getUnit().doubleValue(unit, value);
  }

  /**
   * Parse a value along with its unit from a String representation, and
   * convert it to this unit.
   * @param unitValue The value with a unit in its String
   * representation. If the unit is not specified, then assume it is Pixel.
   * @return The int value resulting from the conversion.
   */
  public static int intValue(String unitValue) {
    return (int)Math.round(doubleValue(unitValue));
  }

  /**
   * Parse a value along with its unit from a String representation, and
   * convert it to this unit.
   * @param unitValue The value with a unit in its String
   * representation. If the unit is not specified, then assume it is Pixel.
   * @return The double value resulting from the conversion.
   */
  public static double doubleValue(String unitValue) {
    return doubleValue(Pixel.getUnit(), Unit.parse(unitValue));
  }

}
